@@ -26,6 +26,7 @@
 
 enum {
     TCP,
+    TCP_PI,
     RTU
 };
 
@@ -37,21 +38,22 @@ int main(int argc, char *argv[])
     modbus_t *ctx;
     int i;
     uint8_t value;
-    int address;
     int nb_points;
     int rc;
     float real;
-    struct timeval timeout_begin_old;
-    struct timeval timeout_begin_new;
+    struct timeval old_response_timeout;
+    struct timeval response_timeout;
     int use_backend;
 
     if (argc > 1) {
         if (strcmp(argv[1], "tcp") == 0) {
             use_backend = TCP;
+	} else if (strcmp(argv[1], "tcppi") == 0) {
+            use_backend = TCP_PI;
         } else if (strcmp(argv[1], "rtu") == 0) {
             use_backend = RTU;
         } else {
-            printf("Usage:\n  %s [tcp|rtu] - Modbus client for unit testing\n\n", argv[0]);
+            printf("Usage:\n  %s [tcp|tcppi|rtu] - Modbus client for unit testing\n\n", argv[0]);
             exit(1);
         }
     } else {
@@ -61,6 +63,8 @@ int main(int argc, char *argv[])
 
     if (use_backend == TCP) {
         ctx = modbus_new_tcp("127.0.0.1", 1502);
+    } else if (use_backend == TCP_PI) {
+        ctx = modbus_new_tcp_pi("::1", "1502");
     } else {
         ctx = modbus_new_rtu("/dev/ttyUSB1", 115200, 'N', 8, 1);
     }
@@ -69,6 +73,9 @@ int main(int argc, char *argv[])
         return -1;
     }
     modbus_set_debug(ctx, TRUE);
+    modbus_set_error_recovery(ctx,
+                              MODBUS_ERROR_RECOVERY_LINK |
+                              MODBUS_ERROR_RECOVERY_PROTOCOL);
 
     if (use_backend == RTU) {
           modbus_set_slave(ctx, SERVER_ID);
@@ -82,15 +89,13 @@ int main(int argc, char *argv[])
     }
 
     /* Allocate and initialize the memory to store the bits */
-    nb_points = (UT_BITS_NB_POINTS > UT_INPUT_BITS_NB_POINTS) ?
-        UT_BITS_NB_POINTS : UT_INPUT_BITS_NB_POINTS;
+    nb_points = (UT_BITS_NB > UT_INPUT_BITS_NB) ? UT_BITS_NB : UT_INPUT_BITS_NB;
     tab_rp_bits = (uint8_t *) malloc(nb_points * sizeof(uint8_t));
     memset(tab_rp_bits, 0, nb_points * sizeof(uint8_t));
 
     /* Allocate and initialize the memory to store the registers */
-    nb_points = (UT_REGISTERS_NB_POINTS >
-                 UT_INPUT_REGISTERS_NB_POINTS) ?
-        UT_REGISTERS_NB_POINTS : UT_INPUT_REGISTERS_NB_POINTS;
+    nb_points = (UT_REGISTERS_NB > UT_INPUT_REGISTERS_NB) ?
+        UT_REGISTERS_NB : UT_INPUT_REGISTERS_NB;
     tab_rp_registers = (uint16_t *) malloc(nb_points * sizeof(uint16_t));
     memset(tab_rp_registers, 0, nb_points * sizeof(uint16_t));
 
@@ -126,14 +131,13 @@ int main(int argc, char *argv[])
 
     /* Multiple bits */
     {
-        uint8_t tab_value[UT_BITS_NB_POINTS];
+        uint8_t tab_value[UT_BITS_NB];
 
-        modbus_set_bits_from_bytes(tab_value, 0, UT_BITS_NB_POINTS,
-                                   UT_BITS_TAB);
+        modbus_set_bits_from_bytes(tab_value, 0, UT_BITS_NB, UT_BITS_TAB);
         rc = modbus_write_bits(ctx, UT_BITS_ADDRESS,
-                               UT_BITS_NB_POINTS, tab_value);
+                               UT_BITS_NB, tab_value);
         printf("1/2 modbus_write_bits: ");
-        if (rc == UT_BITS_NB_POINTS) {
+        if (rc == UT_BITS_NB) {
             printf("OK\n");
         } else {
             printf("FAILED\n");
@@ -141,24 +145,21 @@ int main(int argc, char *argv[])
         }
     }
 
-    rc = modbus_read_bits(ctx, UT_BITS_ADDRESS,
-                          UT_BITS_NB_POINTS, tab_rp_bits);
+    rc = modbus_read_bits(ctx, UT_BITS_ADDRESS, UT_BITS_NB, tab_rp_bits);
     printf("2/2 modbus_read_bits: ");
-    if (rc != UT_BITS_NB_POINTS) {
+    if (rc != UT_BITS_NB) {
         printf("FAILED (nb points %d)\n", rc);
         goto close;
     }
 
     i = 0;
-    address = UT_BITS_ADDRESS;
-    nb_points = UT_BITS_NB_POINTS;
+    nb_points = UT_BITS_NB;
     while (nb_points > 0) {
         int nb_bits = (nb_points > 8) ? 8 : nb_points;
 
         value = modbus_get_byte_from_bits(tab_rp_bits, i*8, nb_bits);
         if (value != UT_BITS_TAB[i]) {
-            printf("FAILED (%0X != %0X)\n",
-                   value, UT_BITS_TAB[i]);
+            printf("FAILED (%0X != %0X)\n", value, UT_BITS_TAB[i]);
             goto close;
         }
 
@@ -170,24 +171,22 @@ int main(int argc, char *argv[])
 
     /** DISCRETE INPUTS **/
     rc = modbus_read_input_bits(ctx, UT_INPUT_BITS_ADDRESS,
-                                UT_INPUT_BITS_NB_POINTS, tab_rp_bits);
+                                UT_INPUT_BITS_NB, tab_rp_bits);
     printf("1/1 modbus_read_input_bits: ");
 
-    if (rc != UT_INPUT_BITS_NB_POINTS) {
+    if (rc != UT_INPUT_BITS_NB) {
         printf("FAILED (nb points %d)\n", rc);
         goto close;
     }
 
     i = 0;
-    address = UT_INPUT_BITS_ADDRESS;
-    nb_points = UT_INPUT_BITS_NB_POINTS;
+    nb_points = UT_INPUT_BITS_NB;
     while (nb_points > 0) {
         int nb_bits = (nb_points > 8) ? 8 : nb_points;
 
         value = modbus_get_byte_from_bits(tab_rp_bits, i*8, nb_bits);
         if (value != UT_INPUT_BITS_TAB[i]) {
-            printf("FAILED (%0X != %0X)\n",
-                   value, UT_INPUT_BITS_TAB[i]);
+            printf("FAILED (%0X != %0X)\n", value, UT_INPUT_BITS_TAB[i]);
             goto close;
         }
 
@@ -226,10 +225,9 @@ int main(int argc, char *argv[])
 
     /* Many registers */
     rc = modbus_write_registers(ctx, UT_REGISTERS_ADDRESS,
-                                UT_REGISTERS_NB_POINTS,
-                                UT_REGISTERS_TAB);
+                                UT_REGISTERS_NB, UT_REGISTERS_TAB);
     printf("1/5 modbus_write_registers: ");
-    if (rc == UT_REGISTERS_NB_POINTS) {
+    if (rc == UT_REGISTERS_NB) {
         printf("OK\n");
     } else {
         printf("FAILED\n");
@@ -237,15 +235,14 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB, tab_rp_registers);
     printf("2/5 modbus_read_registers: ");
-    if (rc != UT_REGISTERS_NB_POINTS) {
+    if (rc != UT_REGISTERS_NB) {
         printf("FAILED (nb points %d)\n", rc);
         goto close;
     }
 
-    for (i=0; i < UT_REGISTERS_NB_POINTS; i++) {
+    for (i=0; i < UT_REGISTERS_NB; i++) {
         if (tab_rp_registers[i] != UT_REGISTERS_TAB[i]) {
             printf("FAILED (%0X != %0X)\n",
                    tab_rp_registers[i],
@@ -264,24 +261,23 @@ int main(int argc, char *argv[])
     }
     printf("OK\n");
 
-    nb_points = (UT_REGISTERS_NB_POINTS >
-                 UT_INPUT_REGISTERS_NB_POINTS) ?
-        UT_REGISTERS_NB_POINTS : UT_INPUT_REGISTERS_NB_POINTS;
+    nb_points = (UT_REGISTERS_NB >
+                 UT_INPUT_REGISTERS_NB) ?
+        UT_REGISTERS_NB : UT_INPUT_REGISTERS_NB;
     memset(tab_rp_registers, 0, nb_points * sizeof(uint16_t));
 
     /* Write registers to zero from tab_rp_registers and store read registers
        into tab_rp_registers. So the read registers must set to 0, except the
        first one because there is an offset of 1 register on write. */
-    rc = modbus_read_and_write_registers(ctx,
-                                         UT_REGISTERS_ADDRESS,
-                                         UT_REGISTERS_NB_POINTS,
+    rc = modbus_write_and_read_registers(ctx,
+                                         UT_REGISTERS_ADDRESS + 1, UT_REGISTERS_NB - 1,
                                          tab_rp_registers,
-                                         UT_REGISTERS_ADDRESS + 1,
-                                         UT_REGISTERS_NB_POINTS - 1,
+                                         UT_REGISTERS_ADDRESS,
+                                         UT_REGISTERS_NB,
                                          tab_rp_registers);
-    printf("4/5 modbus_read_and_write_registers: ");
-    if (rc != UT_REGISTERS_NB_POINTS) {
-        printf("FAILED (nb points %d != %d)\n", rc, UT_REGISTERS_NB_POINTS);
+    printf("4/5 modbus_write_and_read_registers: ");
+    if (rc != UT_REGISTERS_NB) {
+        printf("FAILED (nb points %d != %d)\n", rc, UT_REGISTERS_NB);
         goto close;
     }
 
@@ -290,7 +286,7 @@ int main(int argc, char *argv[])
                tab_rp_registers[0], UT_REGISTERS_TAB[0]);
     }
 
-    for (i=1; i < UT_REGISTERS_NB_POINTS; i++) {
+    for (i=1; i < UT_REGISTERS_NB; i++) {
         if (tab_rp_registers[i] != 0) {
             printf("FAILED (%0X != %0X)\n",
                    tab_rp_registers[i], 0);
@@ -304,15 +300,15 @@ int main(int argc, char *argv[])
 
     /** INPUT REGISTERS **/
     rc = modbus_read_input_registers(ctx, UT_INPUT_REGISTERS_ADDRESS,
-                                     UT_INPUT_REGISTERS_NB_POINTS,
+                                     UT_INPUT_REGISTERS_NB,
                                      tab_rp_registers);
     printf("1/1 modbus_read_input_registers: ");
-    if (rc != UT_INPUT_REGISTERS_NB_POINTS) {
+    if (rc != UT_INPUT_REGISTERS_NB) {
         printf("FAILED (nb points %d)\n", rc);
         goto close;
     }
 
-    for (i=0; i < UT_INPUT_REGISTERS_NB_POINTS; i++) {
+    for (i=0; i < UT_INPUT_REGISTERS_NB; i++) {
         if (tab_rp_registers[i] != UT_INPUT_REGISTERS_TAB[i]) {
             printf("FAILED (%0X != %0X)\n",
                    tab_rp_registers[i], UT_INPUT_REGISTERS_TAB[i]);
@@ -352,8 +348,7 @@ int main(int argc, char *argv[])
      * the addresses are not valid. */
 
     rc = modbus_read_bits(ctx, UT_BITS_ADDRESS,
-                          UT_BITS_NB_POINTS + 1,
-                          tab_rp_bits);
+                          UT_BITS_NB + 1, tab_rp_bits);
     printf("* modbus_read_bits: ");
     if (rc == -1 && errno == EMBXILADD) {
         printf("OK\n");
@@ -363,8 +358,7 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_input_bits(ctx, UT_INPUT_BITS_ADDRESS,
-                                UT_INPUT_BITS_NB_POINTS + 1,
-                                tab_rp_bits);
+                                UT_INPUT_BITS_NB + 1, tab_rp_bits);
     printf("* modbus_read_input_bits: ");
     if (rc == -1 && errno == EMBXILADD)
         printf("OK\n");
@@ -374,8 +368,7 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS + 1,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB + 1, tab_rp_registers);
     printf("* modbus_read_registers: ");
     if (rc == -1 && errno == EMBXILADD)
         printf("OK\n");
@@ -385,7 +378,7 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_input_registers(ctx, UT_INPUT_REGISTERS_ADDRESS,
-                                     UT_INPUT_REGISTERS_NB_POINTS + 1,
+                                     UT_INPUT_REGISTERS_NB + 1,
                                      tab_rp_registers);
     printf("* modbus_read_input_registers: ");
     if (rc == -1 && errno == EMBXILADD)
@@ -395,7 +388,7 @@ int main(int argc, char *argv[])
         goto close;
     }
 
-    rc = modbus_write_bit(ctx, UT_BITS_ADDRESS + UT_BITS_NB_POINTS, ON);
+    rc = modbus_write_bit(ctx, UT_BITS_ADDRESS + UT_BITS_NB, ON);
     printf("* modbus_write_bit: ");
     if (rc == -1 && errno == EMBXILADD) {
         printf("OK\n");
@@ -404,9 +397,8 @@ int main(int argc, char *argv[])
         goto close;
     }
 
-    rc = modbus_write_bits(ctx, UT_BITS_ADDRESS + UT_BITS_NB_POINTS,
-                           UT_BITS_NB_POINTS,
-                           tab_rp_bits);
+    rc = modbus_write_bits(ctx, UT_BITS_ADDRESS + UT_BITS_NB,
+                           UT_BITS_NB, tab_rp_bits);
     printf("* modbus_write_coils: ");
     if (rc == -1 && errno == EMBXILADD) {
         printf("OK\n");
@@ -415,10 +407,8 @@ int main(int argc, char *argv[])
         goto close;
     }
 
-    rc = modbus_write_registers(ctx, UT_REGISTERS_ADDRESS +
-                                UT_REGISTERS_NB_POINTS,
-                                UT_REGISTERS_NB_POINTS,
-                                tab_rp_registers);
+    rc = modbus_write_registers(ctx, UT_REGISTERS_ADDRESS + UT_REGISTERS_NB,
+                                UT_REGISTERS_NB, tab_rp_registers);
     printf("* modbus_write_registers: ");
     if (rc == -1 && errno == EMBXILADD) {
         printf("OK\n");
@@ -431,8 +421,8 @@ int main(int argc, char *argv[])
     /** TOO MANY DATA **/
     printf("\nTEST TOO MANY DATA ERROR:\n");
 
-    rc = modbus_read_bits(ctx, UT_BITS_ADDRESS, MODBUS_MAX_READ_BITS + 1,
-                          tab_rp_bits);
+    rc = modbus_read_bits(ctx, UT_BITS_ADDRESS,
+                          MODBUS_MAX_READ_BITS + 1, tab_rp_bits);
     printf("* modbus_read_bits: ");
     if (rc == -1 && errno == EMBMDATA) {
         printf("OK\n");
@@ -442,8 +432,7 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_input_bits(ctx, UT_INPUT_BITS_ADDRESS,
-                                MODBUS_MAX_READ_BITS + 1,
-                                tab_rp_bits);
+                                MODBUS_MAX_READ_BITS + 1, tab_rp_bits);
     printf("* modbus_read_input_bits: ");
     if (rc == -1 && errno == EMBMDATA) {
         printf("OK\n");
@@ -475,8 +464,7 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_write_bits(ctx, UT_BITS_ADDRESS,
-                           MODBUS_MAX_WRITE_BITS + 1,
-                           tab_rp_bits);
+                           MODBUS_MAX_WRITE_BITS + 1, tab_rp_bits);
     printf("* modbus_write_bits: ");
     if (rc == -1 && errno == EMBMDATA) {
         printf("OK\n");
@@ -498,13 +486,16 @@ int main(int argc, char *argv[])
 
     /** SLAVE REPLY **/
     printf("\nTEST SLAVE REPLY:\n");
-    modbus_set_slave(ctx, 18);
+    modbus_set_slave(ctx, INVALID_SERVER_ID);
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB, tab_rp_registers);
     if (use_backend == RTU) {
+        const int RAW_REQ_LENGTH = 6;
+        uint8_t raw_req[] = { INVALID_SERVER_ID, 0x03, 0x00, 0x01, 0xFF, 0xFF };
+        uint8_t rsp[MODBUS_TCP_MAX_ADU_LENGTH];
+
         /* No response in RTU mode */
-        printf("1/4 No response from slave %d: ", 18);
+        printf("1/4-A No response from slave %d: ", INVALID_SERVER_ID);
 
         if (rc == -1 && errno == ETIMEDOUT) {
             printf("OK\n");
@@ -512,11 +503,27 @@ int main(int argc, char *argv[])
             printf("FAILED\n");
             goto close;
         }
+
+        /* Send an invalid query with a wrong slave ID */
+        modbus_send_raw_request(ctx, raw_req,
+                                RAW_REQ_LENGTH * sizeof(uint8_t));
+        rc = modbus_receive_confirmation(ctx, rsp);
+
+        printf("1/4-B No response from slave %d with invalid request: ",
+               INVALID_SERVER_ID);
+
+        if (rc == -1 && errno == ETIMEDOUT) {
+            printf("OK\n");
+        } else {
+            printf("FAILED (%d)\n", rc);
+            goto close;
+        }
+
     } else {
         /* Response in TCP mode */
         printf("1/4 Response from slave %d: ", 18);
 
-        if (rc == UT_REGISTERS_NB_POINTS) {
+        if (rc == UT_REGISTERS_NB) {
             printf("OK\n");
         } else {
             printf("FAILED\n");
@@ -531,10 +538,9 @@ int main(int argc, char *argv[])
     }
 
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB, tab_rp_registers);
     printf("2/4 Reply after a broadcast query: ");
-    if (rc == UT_REGISTERS_NB_POINTS) {
+    if (rc == UT_REGISTERS_NB) {
         printf("OK\n");
     } else {
         printf("FAILED\n");
@@ -556,49 +562,62 @@ int main(int argc, char *argv[])
         goto close;
     }
 
-    if (((use_backend == RTU) && (tab_rp_bits[0] == SERVER_ID))
-        || tab_rp_bits[0] == 0xFF) {
-        printf("OK\n");
+    /* Slave ID is an arbitraty number for libmodbus */
+    if (rc > 0) {
+        printf("OK Slave ID is %d\n", tab_rp_bits[0]);
     } else {
         printf("FAILED\n");
         goto close;
     }
 
+    /* Run status indicator */
+    if (rc > 1 && tab_rp_bits[1] == 0xFF) {
+        printf("OK Run Status Indicator is %s\n", tab_rp_bits[1] ? "ON" : "OFF");
+    } else {
+        printf("FAILED\n");
+        goto close;
+    }
+
+    /* Print additional data as string */
+    if (rc > 2) {
+        printf("Additional data: ");
+        for (i=2; i < rc; i++) {
+            printf("%c", tab_rp_bits[i]);
+        }
+        printf("\n");
+    }
+
     /* Save original timeout */
-    modbus_get_timeout_begin(ctx, &timeout_begin_old);
+    modbus_get_response_timeout(ctx, &old_response_timeout);
 
     /* Define a new and too short timeout */
-    timeout_begin_new.tv_sec = 0;
-    timeout_begin_new.tv_usec = 0;
-    modbus_set_timeout_begin(ctx, &timeout_begin_new);
+    response_timeout.tv_sec = 0;
+    response_timeout.tv_usec = 0;
+    modbus_set_response_timeout(ctx, &response_timeout);
 
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB, tab_rp_registers);
     printf("4/4 Too short timeout: ");
     if (rc == -1 && errno == ETIMEDOUT) {
         printf("OK\n");
     } else {
-        printf("FAILED\n");
-        goto close;
+        printf("FAILED (can fail on slow systems or Windows)\n");
     }
 
     /* Restore original timeout */
-    modbus_set_timeout_begin(ctx, &timeout_begin_old);
+    modbus_set_response_timeout(ctx, &old_response_timeout);
 
-    /* Wait for data before flushing */
-    usleep(100);
-    modbus_flush(ctx);
+    /* A wait and flush operation is done by the error recovery code of
+     * libmodbus */
 
     /** BAD RESPONSE **/
     printf("\nTEST BAD RESPONSE ERROR:\n");
 
     /* Allocate only the required space */
     tab_rp_registers_bad = (uint16_t *) malloc(
-        UT_REGISTERS_NB_POINTS_SPECIAL * sizeof(uint16_t));
+        UT_REGISTERS_NB_SPECIAL * sizeof(uint16_t));
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS,
-                               UT_REGISTERS_NB_POINTS_SPECIAL,
-                               tab_rp_registers_bad);
+                               UT_REGISTERS_NB_SPECIAL, tab_rp_registers_bad);
     printf("* modbus_read_registers: ");
     if (rc == -1 && errno == EMBBADDATA) {
         printf("OK\n");
@@ -613,8 +632,7 @@ int main(int argc, char *argv[])
     printf("\nTEST MANUAL EXCEPTION:\n");
 
     rc = modbus_read_registers(ctx, UT_REGISTERS_ADDRESS_SPECIAL,
-                               UT_REGISTERS_NB_POINTS,
-                               tab_rp_registers);
+                               UT_REGISTERS_NB, tab_rp_registers);
     printf("* modbus_read_registers at special address: ");
     if (rc == -1 && errno == EMBXSBUSY) {
         printf("OK\n");
@@ -623,6 +641,40 @@ int main(int argc, char *argv[])
         goto close;
     }
 
+    /** RAW REQUEST */
+    printf("\nTEST RAW REQUEST:\n");
+    {
+        const int RAW_REQ_LENGTH = 6;
+        uint8_t raw_req[] = { (use_backend == RTU) ? SERVER_ID : 0xFF,
+                              0x03, 0x00, 0x01, 0x0, 0x05 };
+        int req_length;
+        uint8_t rsp[MODBUS_TCP_MAX_ADU_LENGTH];
+
+        req_length = modbus_send_raw_request(ctx, raw_req,
+                                             RAW_REQ_LENGTH * sizeof(uint8_t));
+
+        printf("* modbus_send_raw_request: ");
+        if ((use_backend == RTU && req_length == (RAW_REQ_LENGTH + 2)) ||
+            ((use_backend == TCP || use_backend == TCP_PI) &&
+             req_length == (RAW_REQ_LENGTH + 6))) {
+            printf("OK\n");
+        } else {
+            printf("FAILED (%d)\n", req_length);
+            goto close;
+        }
+
+        printf("* modbus_receive_confirmation: ");
+        rc  = modbus_receive_confirmation(ctx, rsp);
+        if ((use_backend == RTU && rc == 15) ||
+            ((use_backend == TCP || use_backend == TCP_PI) &&
+             rc == 19)) {
+            printf("OK\n");
+        } else {
+            printf("FAILED (%d)\n", rc);
+            goto close;
+        }
+    }
+
     printf("\nALL TESTS PASS WITH SUCCESS.\n");
 
 close:
